class DataMappingException(Exception):
    pass


class StixValidationException(Exception):
    pass


class TranslationResultException(Exception):
    pass


class LoadJsonResultsException(Exception):
    def __str__(self):
        return "Error when loading the JSON results from the data source"


class UnsupportedDataSourceException(Exception):
    pass



from requests.exceptions import SSLError, ConnectionError
from enum import Enum
import importlib
import traceback
#from ..utils.error_mapper_base import ErrorMapperBase
from stix_shifter_utils.utils.error_mapper_base import ErrorMapperBase
import collections
import json

class ErrorCode(Enum):
    TRANSLATION_NOTIMPLEMENTED_MODE = 'not_implemented'
    TRANSLATION_MODULE_DEFAULT_ERROR =  'invalid_parameter'
    TRANSLATION_MAPPING_ERROR = 'mapping_error'
    TRANSLATION_STIX_VALIDATION = 'invalid_parameter'
    TRANSLATION_NOTSUPPORTED = 'invalid_parameter'
    TRANSLATION_RESULT = 'mapping_error'

    TRANSMISSION_UNKNOWN = 'unknown'
    TRANSMISSION_CONNECT = 'service_unavailable'
    TRANSMISSION_AUTH_SSL = 'authentication_fail'
    TRANSMISSION_AUTH_CREDENTIALS = 'authentication_fail'
    TRANSMISSION_MODULE_DEFAULT_ERROR = 'unknown'
    TRANSMISSION_QUERY_PARSING_ERROR = 'invalid_query'
    TRANSMISSION_QUERY_LOGICAL_ERROR = 'invalid_query'
    TRANSMISSION_RESPONSE_EMPTY_RESULT = 'no_results'
    TRANSMISSION_SEARCH_DOES_NOT_EXISTS = 'no_results'
    TRANSMISSION_INVALID_PARAMETER = 'invalid_parameter'
    TRANSMISSION_REMOTE_SYSTEM_IS_UNAVAILABLE = 'service_unavailable'


class ErrorResponder():

    @staticmethod
    def get_struct_item(message_struct, message_path):
        # "+isFailure=True" means the current item is a list and the new item will be a list containing items with the field 'Failure' equal 'True'
        # '~result' means the current item is a list and new item will be a list containing specified field ('result') values
        # document it: '+' and '~'
        if message_struct is not None and message_path is not None:
            if (isinstance(message_struct, collections.Mapping) or type(message_struct).__name__=='list'):
                struct = message_struct.copy()
                for i in message_path:
                    if (isinstance(struct, collections.Mapping) and i in struct) or (type(struct).__name__=='list' and isinstance(i, int) and i < len(struct)):
                        struct = struct[i]
                        if struct is None:
                            break
                    elif i[:1] == '+' and isinstance(struct, list):
                        key, value = i[1:].split('=')
                        filtered_struct = list(filter(lambda item: (key in item and str(item[key])==str(value)), struct))
                        struct = filtered_struct
                    elif i[:1] == '~' and isinstance(struct, list):
                        key = i[1:]
                        filtered_struct = set()
                        for s in struct:
                            filtered_struct.add(s[key])
                        struct = list(filtered_struct)
                    else:
                        break
                return struct
            else:
                return message_struct

    @staticmethod
    def fill_error(return_object, message_struct=None, message_path=None, message=None, error=None):
        return_object['success'] = False
        error_code = ErrorCode.TRANSMISSION_UNKNOWN
        
        if message is None:
            message = ''

        struct_item = ErrorResponder.get_struct_item(message_struct, message_path)
        if struct_item is not None:
            if len(message) > 0:
                message += ';'
            if (isinstance(struct_item, list)):
                struct_item = json.dumps(struct_item)
            message += str(struct_item)
        error_msg = ''
        if error is not None:
            str_error = str(error)
            #TODO replace with logger + stacktrace it to logger
            print("error occurred: " + str_error)
            if isinstance(error, SSLError):
                error_code = ErrorCode.TRANSMISSION_AUTH_SSL
                error_msg = 'Wrong certificate: ' + str_error
            elif isinstance(error, ConnectionError):
                error_code = ErrorCode.TRANSMISSION_CONNECT
                error_msg = 'Connection error: ' + str_error
            else:
                error_msg = str(error)

            if len(error_msg) > 0:
                if len(message) > 0:
                    message += '; '
                message += error_msg

        if message is not None and len(message)>0:
            if error_code.value == ErrorCode.TRANSMISSION_UNKNOWN.value:
                if 'uthenticat' in message or 'uthoriz' in message:
                    error_code = ErrorCode.TRANSMISSION_AUTH_CREDENTIALS
                elif 'query_syntax_error' in message:
                    error_code = ErrorCode.TRANSMISSION_QUERY_PARSING_ERROR
            return_object['error'] = str(message)
        ErrorMapperBase.set_error_code(return_object, error_code.value)
        if error_code == ErrorCode.TRANSMISSION_UNKNOWN:
            ErrorResponder.call_module_error_mapper(message_struct, return_object)

    @staticmethod
    def call_module_error_mapper(json_data, return_object):
        caller_path_list = traceback.extract_stack()[-3].filename.split('/')

        if 'stix_translation.py' in caller_path_list[-1]:
            module_path = 'stix_shifter_utils.stix_translation.stix_translation_error_mapper'
        else:
            caller_module_name = caller_path_list[-3:-1]
            module_path = 'stix_shifter_modules.' + caller_module_name[0] + '.'  + caller_module_name[1] + '.' + caller_module_name[0] + '_error_mapper'

        # path_start_position = ErrorResponder.rindex(caller_path_list, 'stix-shifter')
        # module_path = 'stix_shifter_modules.' + caller_module_name[0] + '.'  + caller_module_name[1] + '.' + caller_module_name[0] + '_error_mapper'
        # module_path = '.'.join(caller_path_list[path_start_position: -1]) + '.' + caller_module_name + '_error_mapper'
        try:
            module = importlib.import_module(module_path)
            if json_data is not None:
                module.ErrorMapper.set_error_code(json_data, return_object)
            else:
                ErrorMapperBase.set_error_code(return_object, module.ErrorMapper.DEFAULT_ERROR)
        except ModuleNotFoundError:
            pass


    @staticmethod
    def rindex(mylist, myvalue):
        return len(mylist) - mylist[::-1].index(myvalue) - 1

    @staticmethod
    def is_plain_string(s):
        return isinstance(s, str) and not s.startswith('<?') and not s.startswith('{')

    @staticmethod
    def is_json_string(s):
        return isinstance(s, str) and s.startswith('{')



from abc import ABCMeta, abstractmethod


class BaseResultTranslator(object, metaclass=ABCMeta):

    def __init__(self, default_mapping_file_path=None, callback=None):
        self.default_mapping_file_path = default_mapping_file_path
        self.callback = callback

    @abstractmethod
    def translate_results(self, data_source, data, options, mapping=None):
        """
        Translates data into STIX results based on a mapping file
        :param data_source: STIX identity object representing a data source
        :type data_source: str
        :param data: data to translate into STIX format
        :type data: str
        :param mapping: The mapping file path to use as instructions on how to translate the given data to STIX. This should default to something if it hasn't been passed in
        :type mapping: str (filepath)
        :return: translated STIX formatted results
        :rtype: str
        """
        # if translating some datasource to STIX results...
        pass



from abc import ABCMeta, abstractmethod


class BaseQueryConnector(object, metaclass=ABCMeta):
    @abstractmethod
    def create_query_connection(self, query):
        """
        Creates a connection to the specified datasource to send a query

        Args:
            query (str): The datasource query.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    search_id (str): query ID
                    error (str): error message (when success=False)
        """
        pass



# Generated from STIXPattern.g4 by ANTLR 4.7.1
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .STIXPatternParser import STIXPatternParser
else:
    from STIXPatternParser import STIXPatternParser

# This class defines a complete listener for a parse tree produced by STIXPatternParser.
class STIXPatternListener(ParseTreeListener):

    # Enter a parse tree produced by STIXPatternParser#pattern.
    def enterPattern(self, ctx:STIXPatternParser.PatternContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#pattern.
    def exitPattern(self, ctx:STIXPatternParser.PatternContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressions.
    def enterObservationExpressions(self, ctx:STIXPatternParser.ObservationExpressionsContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressions.
    def exitObservationExpressions(self, ctx:STIXPatternParser.ObservationExpressionsContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionOr.
    def enterObservationExpressionOr(self, ctx:STIXPatternParser.ObservationExpressionOrContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionOr.
    def exitObservationExpressionOr(self, ctx:STIXPatternParser.ObservationExpressionOrContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionAnd.
    def enterObservationExpressionAnd(self, ctx:STIXPatternParser.ObservationExpressionAndContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionAnd.
    def exitObservationExpressionAnd(self, ctx:STIXPatternParser.ObservationExpressionAndContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionRepeated.
    def enterObservationExpressionRepeated(self, ctx:STIXPatternParser.ObservationExpressionRepeatedContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionRepeated.
    def exitObservationExpressionRepeated(self, ctx:STIXPatternParser.ObservationExpressionRepeatedContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionSimple.
    def enterObservationExpressionSimple(self, ctx:STIXPatternParser.ObservationExpressionSimpleContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionSimple.
    def exitObservationExpressionSimple(self, ctx:STIXPatternParser.ObservationExpressionSimpleContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionCompound.
    def enterObservationExpressionCompound(self, ctx:STIXPatternParser.ObservationExpressionCompoundContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionCompound.
    def exitObservationExpressionCompound(self, ctx:STIXPatternParser.ObservationExpressionCompoundContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionWithin.
    def enterObservationExpressionWithin(self, ctx:STIXPatternParser.ObservationExpressionWithinContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionWithin.
    def exitObservationExpressionWithin(self, ctx:STIXPatternParser.ObservationExpressionWithinContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#observationExpressionStartStop.
    def enterObservationExpressionStartStop(self, ctx:STIXPatternParser.ObservationExpressionStartStopContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#observationExpressionStartStop.
    def exitObservationExpressionStartStop(self, ctx:STIXPatternParser.ObservationExpressionStartStopContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#comparisonExpressionAnd_.
    def enterComparisonExpressionAnd_(self, ctx:STIXPatternParser.ComparisonExpressionAnd_Context):
        pass

    # Exit a parse tree produced by STIXPatternParser#comparisonExpressionAnd_.
    def exitComparisonExpressionAnd_(self, ctx:STIXPatternParser.ComparisonExpressionAnd_Context):
        pass


    # Enter a parse tree produced by STIXPatternParser#comparisonExpressionOred.
    def enterComparisonExpressionOred(self, ctx:STIXPatternParser.ComparisonExpressionOredContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#comparisonExpressionOred.
    def exitComparisonExpressionOred(self, ctx:STIXPatternParser.ComparisonExpressionOredContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#comparisonExpressionAndPropTest.
    def enterComparisonExpressionAndPropTest(self, ctx:STIXPatternParser.ComparisonExpressionAndPropTestContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#comparisonExpressionAndPropTest.
    def exitComparisonExpressionAndPropTest(self, ctx:STIXPatternParser.ComparisonExpressionAndPropTestContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#comparisonExpressionAnded.
    def enterComparisonExpressionAnded(self, ctx:STIXPatternParser.ComparisonExpressionAndedContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#comparisonExpressionAnded.
    def exitComparisonExpressionAnded(self, ctx:STIXPatternParser.ComparisonExpressionAndedContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestEqual.
    def enterPropTestEqual(self, ctx:STIXPatternParser.PropTestEqualContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestEqual.
    def exitPropTestEqual(self, ctx:STIXPatternParser.PropTestEqualContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestOrder.
    def enterPropTestOrder(self, ctx:STIXPatternParser.PropTestOrderContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestOrder.
    def exitPropTestOrder(self, ctx:STIXPatternParser.PropTestOrderContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestSet.
    def enterPropTestSet(self, ctx:STIXPatternParser.PropTestSetContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestSet.
    def exitPropTestSet(self, ctx:STIXPatternParser.PropTestSetContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestLike.
    def enterPropTestLike(self, ctx:STIXPatternParser.PropTestLikeContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestLike.
    def exitPropTestLike(self, ctx:STIXPatternParser.PropTestLikeContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestRegex.
    def enterPropTestRegex(self, ctx:STIXPatternParser.PropTestRegexContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestRegex.
    def exitPropTestRegex(self, ctx:STIXPatternParser.PropTestRegexContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestIsSubset.
    def enterPropTestIsSubset(self, ctx:STIXPatternParser.PropTestIsSubsetContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestIsSubset.
    def exitPropTestIsSubset(self, ctx:STIXPatternParser.PropTestIsSubsetContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestIsSuperset.
    def enterPropTestIsSuperset(self, ctx:STIXPatternParser.PropTestIsSupersetContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestIsSuperset.
    def exitPropTestIsSuperset(self, ctx:STIXPatternParser.PropTestIsSupersetContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#propTestParen.
    def enterPropTestParen(self, ctx:STIXPatternParser.PropTestParenContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#propTestParen.
    def exitPropTestParen(self, ctx:STIXPatternParser.PropTestParenContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#orderingComparator.
    def enterOrderingComparator(self, ctx:STIXPatternParser.OrderingComparatorContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#orderingComparator.
    def exitOrderingComparator(self, ctx:STIXPatternParser.OrderingComparatorContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#stringLiteral.
    def enterStringLiteral(self, ctx:STIXPatternParser.StringLiteralContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#stringLiteral.
    def exitStringLiteral(self, ctx:STIXPatternParser.StringLiteralContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#startStopQualifier.
    def enterStartStopQualifier(self, ctx:STIXPatternParser.StartStopQualifierContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#startStopQualifier.
    def exitStartStopQualifier(self, ctx:STIXPatternParser.StartStopQualifierContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#withinQualifier.
    def enterWithinQualifier(self, ctx:STIXPatternParser.WithinQualifierContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#withinQualifier.
    def exitWithinQualifier(self, ctx:STIXPatternParser.WithinQualifierContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#repeatedQualifier.
    def enterRepeatedQualifier(self, ctx:STIXPatternParser.RepeatedQualifierContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#repeatedQualifier.
    def exitRepeatedQualifier(self, ctx:STIXPatternParser.RepeatedQualifierContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#objectPath.
    def enterObjectPath(self, ctx:STIXPatternParser.ObjectPathContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#objectPath.
    def exitObjectPath(self, ctx:STIXPatternParser.ObjectPathContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#objectType.
    def enterObjectType(self, ctx:STIXPatternParser.ObjectTypeContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#objectType.
    def exitObjectType(self, ctx:STIXPatternParser.ObjectTypeContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#firstPathComponent.
    def enterFirstPathComponent(self, ctx:STIXPatternParser.FirstPathComponentContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#firstPathComponent.
    def exitFirstPathComponent(self, ctx:STIXPatternParser.FirstPathComponentContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#indexPathStep.
    def enterIndexPathStep(self, ctx:STIXPatternParser.IndexPathStepContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#indexPathStep.
    def exitIndexPathStep(self, ctx:STIXPatternParser.IndexPathStepContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#pathStep.
    def enterPathStep(self, ctx:STIXPatternParser.PathStepContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#pathStep.
    def exitPathStep(self, ctx:STIXPatternParser.PathStepContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#keyPathStep.
    def enterKeyPathStep(self, ctx:STIXPatternParser.KeyPathStepContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#keyPathStep.
    def exitKeyPathStep(self, ctx:STIXPatternParser.KeyPathStepContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#setLiteral.
    def enterSetLiteral(self, ctx:STIXPatternParser.SetLiteralContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#setLiteral.
    def exitSetLiteral(self, ctx:STIXPatternParser.SetLiteralContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#primitiveLiteral.
    def enterPrimitiveLiteral(self, ctx:STIXPatternParser.PrimitiveLiteralContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#primitiveLiteral.
    def exitPrimitiveLiteral(self, ctx:STIXPatternParser.PrimitiveLiteralContext):
        pass


    # Enter a parse tree produced by STIXPatternParser#orderableLiteral.
    def enterOrderableLiteral(self, ctx:STIXPatternParser.OrderableLiteralContext):
        pass

    # Exit a parse tree produced by STIXPatternParser#orderableLiteral.
    def exitOrderableLiteral(self, ctx:STIXPatternParser.OrderableLiteralContext):
        pass





import importlib
import traceback
from stix_shifter_utils.utils.module_discovery import dialect_list
import os
from stix_shifter_utils.modules.base.stix_translation.base_translator import BaseTranslator
from stix_shifter_utils.stix_translation.src.json_to_stix.json_to_stix import JSONToStix
from stix_shifter_utils.modules.base.stix_translation.base_query_translator import BaseQueryTranslator
from stix_shifter_utils.modules.base.stix_translation.base_results_translator import BaseResultTranslator
from stix_shifter_utils.modules.base.stix_translation.base_data_mapper import BaseDataMapper
from stix_shifter_utils.modules.base.stix_transmission.base_connector import BaseConnector
from stix_shifter_utils.modules.base.stix_transmission.base_delete_connector import BaseDeleteConnector
from stix_shifter_utils.modules.base.stix_transmission.base_query_connector import BaseQueryConnector
from stix_shifter_utils.modules.base.stix_transmission.base_results_connector import BaseResultsConnector
from stix_shifter_utils.modules.base.stix_transmission.base_status_connector import BaseStatusConnector
from stix_shifter_utils.modules.base.stix_transmission.base_ping_connector import BasePingConnector

import sys, inspect
import functools


class EntryPointBase:

    def __init__(self, options):
        self.__async = True
        stack = traceback.extract_stack()
        self.__connector_module = stack[-2].filename.split('/')[-2]
        self.__dialect_to_query_translator = {}
        self.__dialect_to_results_translator = {}
        self.__dialect_to_data_mapper = {}
        self.__dialects_visible = []
        self.__dialect_default = None
        self.__options = options

        self.__results_connector = None
        self.__status_connector = None
        self.__delete_connector = None
        self.__query_connector = None

    def translation(func):
        @functools.wraps(func)
        def wrapper_func(self, *args, **kwargs):
            self.__check_translation_configured()
            return func(self, *args, **kwargs)
        return wrapper_func

    def __check_translation_configured(self):
        if not self.__dialect_default:
            raise Exception('EntryPoint: default dialect is not configured')

    def transmission(func):
        @functools.wraps(func)
        def wrapper_func(self, *args, **kwargs):
            self.__check_transmission_configured()
            return func(self, *args, **kwargs)
        return wrapper_func

    def __check_transmission_configured(self):
        if not(self.__results_connector and self.__status_connector and self.__delete_connector and self.__query_connector):
            raise Exception('EntryPoint: one of transmission connectors is not configured')

    def add_dialect(self, dialect, data_mapper, query_translator=None, results_translator=None, default=False, default_include=True):
        if not query_translator:
            query_translator = self.create_default_query_translator()
        if not results_translator:
            results_translator = self.create_default_results_translator()
        if not (isinstance(query_translator, (BaseTranslator, BaseQueryTranslator))):
            raise Exception('query_translator is not instance of BaseTranslator or BaseQueryTranslator')
        if not (isinstance(results_translator, (BaseTranslator, BaseResultTranslator))):
            raise Exception('results_translator is not instance of BaseTranslator or BaseResultTranslator')
        if data_mapper:
            if not (isinstance(data_mapper, BaseDataMapper)):
                raise Exception('data_mapper is not instance of BaseDataMapper')
        self.__dialect_to_query_translator[dialect] = query_translator
        self.__dialect_to_results_translator[dialect] = results_translator
        self.__dialect_to_data_mapper[dialect] = data_mapper
        if default:
            self.__dialect_default = dialect
        if default_include:
            self.__dialects_visible.append(dialect)
        return self

    def setup_translation_simple(self, dialect_default, query_translator=None, results_translator=None):
        module = self.__connector_module
        dialects = dialect_list(module)

        datamapper_module = importlib.import_module(
                    "stix_shifter_modules." + module + ".stix_translation.data_mapper")
        for dialect in dialects:
            data_mapper = datamapper_module.DataMapper(self.__options, dialect)
            self.add_dialect(dialect, data_mapper, query_translator=query_translator, results_translator=results_translator, default=(dialect==dialect_default))

    def create_default_query_translator(self):
        module = self.__connector_module
        stixtoquety_module = importlib.import_module(
                    "stix_shifter_modules." + module + ".stix_translation.query_translator")
        query_translator = stixtoquety_module.QueryTranslator()
        return query_translator

    def create_default_results_translator(self):
        module = self.__connector_module
        translation_module = importlib.import_module(
                    "stix_shifter_modules." + module + ".stix_translation")
        basepath = os.path.dirname(translation_module.__file__)
        mapping_filepath = os.path.abspath(os.path.join(basepath, "json", "to_stix_map.json"))
        return JSONToStix(mapping_filepath)

    @translation
    def get_data_mapper(self, dialect=None):
        if dialect:
            return self.__dialect_to_data_mapper[dialect]
        return self.__dialect_to_data_mapper[self.__dialect_default]

    @translation    
    def get_query_translator(self, dialect=None):
        if dialect:
            return self.__dialect_to_query_translator[dialect]
        return self.__dialect_to_query_translator[self.__dialect_default]

    @translation
    def get_results_translator(self, dialect=None):
        if dialect:
            return self.__dialect_to_results_translator[dialect]
        return self.__dialect_to_results_translator[self.__dialect_default]

    @translation
    def transform_query(self, dialect, data, antlr_parsing, options, mapping=None):
        data_model_mapper = self.get_data_mapper(dialect)
        translator = self.get_query_translator(dialect)
        return translator.transform_query(data, antlr_parsing, data_model_mapper, options, mapping)

    @translation
    def translate_results(self, data_source, data, options, mapping=None):
        translator = self.get_results_translator()
        return translator.translate_results(data_source, data, options, mapping)

    @translation
    def get_dialects(self):
        return self.__dialects_visible

    def setup_transmission_basic(self, connector):
        if not isinstance(connector, BaseConnector):
            raise Exception('connector is not instance of BaseConnector')
        self.set_query_connector(connector)
        self.set_status_connector(connector)
        self.set_results_connector(connector)
        self.set_delete_connector(connector)
        self.set_ping_connector(connector)

    def set_query_connector(self, connector):
        if not (isinstance(connector, (BaseConnector, BaseQueryConnector)) or issubclass(connector, BaseConnector)):
            raise Exception('connector is not instance of BaseConnector (or it\'s subclass) or BaseQueryConnector')
        self.__query_connector = connector

    @transmission
    def create_query_connection(self, query):
        return self.__query_connector.create_query_connection(query)

    def set_status_connector(self, connector):
        if not (isinstance(connector, (BaseConnector, BaseStatusConnector)) or issubclass(connector, BaseConnector)):
            raise Exception('connector is not instance of BaseConnector (or it\'s subclass) or BaseStatusConnector')
        self.__status_connector = connector

    @transmission
    def create_status_connection(self, search_id):
        return self.__status_connector.create_status_connection(search_id)
    
    def set_results_connector(self, connector):
        if not isinstance(connector, (BaseConnector, BaseResultsConnector)):
            raise Exception('connector is not instance of BaseConnector or BaseResultsConnector')
        self.__results_connector = connector

    @transmission
    def create_results_connection(self, search_id, offset, length):
        return self.__results_connector.create_results_connection(search_id, offset, length)

    def set_delete_connector(self, connector):
        if not isinstance(connector, (BaseConnector, BaseDeleteConnector)):
            raise Exception('connector is not instance of BaseConnector or BaseDeleteConnector')
        self.__delete_connector = connector

    @transmission
    def delete_query_connection(self, search_id):
        return self.__delete_connector.delete_query_connection(search_id)

    def set_ping_connector(self, connector):
        if not isinstance(connector, (BaseConnector, BasePingConnector)):
            raise Exception('connector is not instance of BaseConnector or BasePingConnector')
        self.__ping_connector = connector

    @transmission
    def ping_connection(self):
        return self.__ping_connector.ping_connection()

    def set_async(self, is_async):
        self.__async = is_async

    def is_async(self):
        return self.__async



# Generated from STIXPattern.g4 by ANTLR 4.7.1
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\67")
        buf.write("\u00f9\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4\23\t\23")
        buf.write("\4\24\t\24\4\25\t\25\3\2\3\2\3\2\3\3\3\3\3\3\3\3\3\3\3")
        buf.write("\3\7\3\64\n\3\f\3\16\3\67\13\3\3\4\3\4\3\4\3\4\3\4\3\4")
        buf.write("\7\4?\n\4\f\4\16\4B\13\4\3\5\3\5\3\5\3\5\3\5\3\5\7\5J")
        buf.write("\n\5\f\5\16\5M\13\5\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3\6\3")
        buf.write("\6\5\6X\n\6\3\6\3\6\3\6\3\6\3\6\3\6\7\6`\n\6\f\6\16\6")
        buf.write("c\13\6\3\7\3\7\3\7\3\7\3\7\3\7\7\7k\n\7\f\7\16\7n\13\7")
        buf.write("\3\b\3\b\3\b\3\b\3\b\3\b\7\bv\n\b\f\b\16\by\13\b\3\t\3")
        buf.write("\t\5\t}\n\t\3\t\3\t\3\t\3\t\3\t\5\t\u0084\n\t\3\t\3\t")
        buf.write("\3\t\3\t\3\t\5\t\u008b\n\t\3\t\3\t\3\t\3\t\3\t\5\t\u0092")
        buf.write("\n\t\3\t\3\t\3\t\3\t\3\t\5\t\u0099\n\t\3\t\3\t\3\t\3\t")
        buf.write("\3\t\5\t\u00a0\n\t\3\t\3\t\3\t\3\t\3\t\5\t\u00a7\n\t\3")
        buf.write("\t\3\t\3\t\3\t\3\t\3\t\3\t\5\t\u00b0\n\t\3\n\3\n\3\13")
        buf.write("\3\13\3\f\3\f\3\f\3\f\3\f\3\r\3\r\3\r\3\r\3\16\3\16\3")
        buf.write("\16\3\16\3\17\3\17\3\17\3\17\5\17\u00c7\n\17\3\20\3\20")
        buf.write("\3\21\3\21\3\22\3\22\3\22\3\22\3\22\3\22\5\22\u00d3\n")
        buf.write("\22\3\22\3\22\7\22\u00d7\n\22\f\22\16\22\u00da\13\22\3")
        buf.write("\23\3\23\3\23\3\23\3\23\3\23\7\23\u00e2\n\23\f\23\16\23")
        buf.write("\u00e5\13\23\3\23\3\23\5\23\u00e9\n\23\3\24\3\24\5\24")
        buf.write("\u00ed\n\24\3\25\3\25\3\25\3\25\3\25\3\25\3\25\3\25\5")
        buf.write("\25\u00f7\n\25\3\25\2\t\4\6\b\n\f\16\"\26\2\4\6\b\n\f")
        buf.write("\16\20\22\24\26\30\32\34\36 \"$&(\2\b\3\2 !\3\2\"%\4\2")
        buf.write("\4\4\6\6\3\2\36\37\4\2\t\t\36\36\4\2\3\4\63\63\2\u0108")
        buf.write("\2*\3\2\2\2\4-\3\2\2\2\68\3\2\2\2\bC\3\2\2\2\nW\3\2\2")
        buf.write("\2\fd\3\2\2\2\16o\3\2\2\2\20\u00af\3\2\2\2\22\u00b1\3")
        buf.write("\2\2\2\24\u00b3\3\2\2\2\26\u00b5\3\2\2\2\30\u00ba\3\2")
        buf.write("\2\2\32\u00be\3\2\2\2\34\u00c2\3\2\2\2\36\u00c8\3\2\2")
        buf.write("\2 \u00ca\3\2\2\2\"\u00d2\3\2\2\2$\u00e8\3\2\2\2&\u00ec")
        buf.write("\3\2\2\2(\u00f6\3\2\2\2*+\5\4\3\2+,\7\2\2\3,\3\3\2\2\2")
        buf.write("-.\b\3\1\2./\5\6\4\2/\65\3\2\2\2\60\61\f\4\2\2\61\62\7")
        buf.write("\17\2\2\62\64\5\4\3\5\63\60\3\2\2\2\64\67\3\2\2\2\65\63")
        buf.write("\3\2\2\2\65\66\3\2\2\2\66\5\3\2\2\2\67\65\3\2\2\289\b")
        buf.write("\4\1\29:\5\b\5\2:@\3\2\2\2;<\f\4\2\2<=\7\r\2\2=?\5\6\4")
        buf.write("\5>;\3\2\2\2?B\3\2\2\2@>\3\2\2\2@A\3\2\2\2A\7\3\2\2\2")
        buf.write("B@\3\2\2\2CD\b\5\1\2DE\5\n\6\2EK\3\2\2\2FG\f\4\2\2GH\7")
        buf.write("\f\2\2HJ\5\b\5\5IF\3\2\2\2JM\3\2\2\2KI\3\2\2\2KL\3\2\2")
        buf.write("\2L\t\3\2\2\2MK\3\2\2\2NO\b\6\1\2OP\7-\2\2PQ\5\f\7\2Q")
        buf.write("R\7,\2\2RX\3\2\2\2ST\7+\2\2TU\5\4\3\2UV\7*\2\2VX\3\2\2")
        buf.write("\2WN\3\2\2\2WS\3\2\2\2Xa\3\2\2\2YZ\f\5\2\2Z`\5\26\f\2")
        buf.write("[\\\f\4\2\2\\`\5\30\r\2]^\f\3\2\2^`\5\32\16\2_Y\3\2\2")
        buf.write("\2_[\3\2\2\2_]\3\2\2\2`c\3\2\2\2a_\3\2\2\2ab\3\2\2\2b")
        buf.write("\13\3\2\2\2ca\3\2\2\2de\b\7\1\2ef\5\16\b\2fl\3\2\2\2g")
        buf.write("h\f\4\2\2hi\7\r\2\2ik\5\f\7\5jg\3\2\2\2kn\3\2\2\2lj\3")
        buf.write("\2\2\2lm\3\2\2\2m\r\3\2\2\2nl\3\2\2\2op\b\b\1\2pq\5\20")
        buf.write("\t\2qw\3\2\2\2rs\f\4\2\2st\7\f\2\2tv\5\16\b\5ur\3\2\2")
        buf.write("\2vy\3\2\2\2wu\3\2\2\2wx\3\2\2\2x\17\3\2\2\2yw\3\2\2\2")
        buf.write("z|\5\34\17\2{}\7\16\2\2|{\3\2\2\2|}\3\2\2\2}~\3\2\2\2")
        buf.write("~\177\t\2\2\2\177\u0080\5&\24\2\u0080\u00b0\3\2\2\2\u0081")
        buf.write("\u0083\5\34\17\2\u0082\u0084\7\16\2\2\u0083\u0082\3\2")
        buf.write("\2\2\u0083\u0084\3\2\2\2\u0084\u0085\3\2\2\2\u0085\u0086")
        buf.write("\t\3\2\2\u0086\u0087\5(\25\2\u0087\u00b0\3\2\2\2\u0088")
        buf.write("\u008a\5\34\17\2\u0089\u008b\7\16\2\2\u008a\u0089\3\2")
        buf.write("\2\2\u008a\u008b\3\2\2\2\u008b\u008c\3\2\2\2\u008c\u008d")
        buf.write("\7\25\2\2\u008d\u008e\5$\23\2\u008e\u00b0\3\2\2\2\u008f")
        buf.write("\u0091\5\34\17\2\u0090\u0092\7\16\2\2\u0091\u0090\3\2")
        buf.write("\2\2\u0091\u0092\3\2\2\2\u0092\u0093\3\2\2\2\u0093\u0094")
        buf.write("\7\20\2\2\u0094\u0095\5\24\13\2\u0095\u00b0\3\2\2\2\u0096")
        buf.write("\u0098\5\34\17\2\u0097\u0099\7\16\2\2\u0098\u0097\3\2")
        buf.write("\2\2\u0098\u0099\3\2\2\2\u0099\u009a\3\2\2\2\u009a\u009b")
        buf.write("\7\21\2\2\u009b\u009c\5\24\13\2\u009c\u00b0\3\2\2\2\u009d")
        buf.write("\u009f\5\34\17\2\u009e\u00a0\7\16\2\2\u009f\u009e\3\2")
        buf.write("\2\2\u009f\u00a0\3\2\2\2\u00a0\u00a1\3\2\2\2\u00a1\u00a2")
        buf.write("\7\23\2\2\u00a2\u00a3\5\24\13\2\u00a3\u00b0\3\2\2\2\u00a4")
        buf.write("\u00a6\5\34\17\2\u00a5\u00a7\7\16\2\2\u00a6\u00a5\3\2")
        buf.write("\2\2\u00a6\u00a7\3\2\2\2\u00a7\u00a8\3\2\2\2\u00a8\u00a9")
        buf.write("\7\22\2\2\u00a9\u00aa\5\24\13\2\u00aa\u00b0\3\2\2\2\u00ab")
        buf.write("\u00ac\7+\2\2\u00ac\u00ad\5\f\7\2\u00ad\u00ae\7*\2\2\u00ae")
        buf.write("\u00b0\3\2\2\2\u00afz\3\2\2\2\u00af\u0081\3\2\2\2\u00af")
        buf.write("\u0088\3\2\2\2\u00af\u008f\3\2\2\2\u00af\u0096\3\2\2\2")
        buf.write("\u00af\u009d\3\2\2\2\u00af\u00a4\3\2\2\2\u00af\u00ab\3")
        buf.write("\2\2\2\u00b0\21\3\2\2\2\u00b1\u00b2\t\3\2\2\u00b2\23\3")
        buf.write("\2\2\2\u00b3\u00b4\7\t\2\2\u00b4\25\3\2\2\2\u00b5\u00b6")
        buf.write("\7\26\2\2\u00b6\u00b7\7\13\2\2\u00b7\u00b8\7\27\2\2\u00b8")
        buf.write("\u00b9\7\13\2\2\u00b9\27\3\2\2\2\u00ba\u00bb\7\33\2\2")
        buf.write("\u00bb\u00bc\t\4\2\2\u00bc\u00bd\7\30\2\2\u00bd\31\3\2")
        buf.write("\2\2\u00be\u00bf\7\34\2\2\u00bf\u00c0\7\4\2\2\u00c0\u00c1")
        buf.write("\7\35\2\2\u00c1\33\3\2\2\2\u00c2\u00c3\5\36\20\2\u00c3")
        buf.write("\u00c4\7\'\2\2\u00c4\u00c6\5 \21\2\u00c5\u00c7\5\"\22")
        buf.write("\2\u00c6\u00c5\3\2\2\2\u00c6\u00c7\3\2\2\2\u00c7\35\3")
        buf.write("\2\2\2\u00c8\u00c9\t\5\2\2\u00c9\37\3\2\2\2\u00ca\u00cb")
        buf.write("\t\6\2\2\u00cb!\3\2\2\2\u00cc\u00cd\b\22\1\2\u00cd\u00ce")
        buf.write("\7(\2\2\u00ce\u00d3\t\6\2\2\u00cf\u00d0\7-\2\2\u00d0\u00d1")
        buf.write("\t\7\2\2\u00d1\u00d3\7,\2\2\u00d2\u00cc\3\2\2\2\u00d2")
        buf.write("\u00cf\3\2\2\2\u00d3\u00d8\3\2\2\2\u00d4\u00d5\f\5\2\2")
        buf.write("\u00d5\u00d7\5\"\22\6\u00d6\u00d4\3\2\2\2\u00d7\u00da")
        buf.write("\3\2\2\2\u00d8\u00d6\3\2\2\2\u00d8\u00d9\3\2\2\2\u00d9")
        buf.write("#\3\2\2\2\u00da\u00d8\3\2\2\2\u00db\u00dc\7+\2\2\u00dc")
        buf.write("\u00e9\7*\2\2\u00dd\u00de\7+\2\2\u00de\u00e3\5&\24\2\u00df")
        buf.write("\u00e0\7)\2\2\u00e0\u00e2\5&\24\2\u00e1\u00df\3\2\2\2")
        buf.write("\u00e2\u00e5\3\2\2\2\u00e3\u00e1\3\2\2\2\u00e3\u00e4\3")
        buf.write("\2\2\2\u00e4\u00e6\3\2\2\2\u00e5\u00e3\3\2\2\2\u00e6\u00e7")
        buf.write("\7*\2\2\u00e7\u00e9\3\2\2\2\u00e8\u00db\3\2\2\2\u00e8")
        buf.write("\u00dd\3\2\2\2\u00e9%\3\2\2\2\u00ea\u00ed\5(\25\2\u00eb")
        buf.write("\u00ed\7\n\2\2\u00ec\u00ea\3\2\2\2\u00ec\u00eb\3\2\2\2")
        buf.write("\u00ed\'\3\2\2\2\u00ee\u00f7\7\4\2\2\u00ef\u00f7\7\3\2")
        buf.write("\2\u00f0\u00f7\7\6\2\2\u00f1\u00f7\7\5\2\2\u00f2\u00f7")
        buf.write("\5\24\13\2\u00f3\u00f7\7\b\2\2\u00f4\u00f7\7\7\2\2\u00f5")
        buf.write("\u00f7\7\13\2\2\u00f6\u00ee\3\2\2\2\u00f6\u00ef\3\2\2")
        buf.write("\2\u00f6\u00f0\3\2\2\2\u00f6\u00f1\3\2\2\2\u00f6\u00f2")
        buf.write("\3\2\2\2\u00f6\u00f3\3\2\2\2\u00f6\u00f4\3\2\2\2\u00f6")
        buf.write("\u00f5\3\2\2\2\u00f7)\3\2\2\2\31\65@KW_alw|\u0083\u008a")
        buf.write("\u0091\u0098\u009f\u00a6\u00af\u00c6\u00d2\u00d8\u00e3")
        buf.write("\u00e8\u00ec\u00f6")
        return buf.getvalue()


class STIXPatternParser ( Parser ):

    grammarFileName = "STIXPattern.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "<INVALID>", "<INVALID>", "'AND'", "'OR'", "'NOT'", 
                     "'FOLLOWEDBY'", "'LIKE'", "'MATCHES'", "'ISSUPERSET'", 
                     "'ISSUBSET'", "'LAST'", "'IN'", "'START'", "'STOP'", 
                     "'SECONDS'", "'true'", "'false'", "'WITHIN'", "'REPEATS'", 
                     "'TIMES'", "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                     "'<'", "'<='", "'>'", "'>='", "'''", "':'", "'.'", 
                     "','", "')'", "'('", "']'", "'['", "'+'", "<INVALID>", 
                     "'-'", "'^'", "'/'", "'*'" ]

    symbolicNames = [ "<INVALID>", "IntNegLiteral", "IntPosLiteral", "FloatNegLiteral", 
                      "FloatPosLiteral", "HexLiteral", "BinaryLiteral", 
                      "StringLiteral", "BoolLiteral", "TimestampLiteral", 
                      "AND", "OR", "NOT", "FOLLOWEDBY", "LIKE", "MATCHES", 
                      "ISSUPERSET", "ISSUBSET", "LAST", "IN", "START", "STOP", 
                      "SECONDS", "TRUE", "FALSE", "WITHIN", "REPEATS", "TIMES", 
                      "IdentifierWithoutHyphen", "IdentifierWithHyphen", 
                      "EQ", "NEQ", "LT", "LE", "GT", "GE", "QUOTE", "COLON", 
                      "DOT", "COMMA", "RPAREN", "LPAREN", "RBRACK", "LBRACK", 
                      "PLUS", "HYPHEN", "MINUS", "POWER_OP", "DIVIDE", "ASTERISK", 
                      "WS", "COMMENT", "LINE_COMMENT", "InvalidCharacter" ]

    RULE_pattern = 0
    RULE_observationExpressions = 1
    RULE_observationExpressionOr = 2
    RULE_observationExpressionAnd = 3
    RULE_observationExpression = 4
    RULE_comparisonExpression = 5
    RULE_comparisonExpressionAnd = 6
    RULE_propTest = 7
    RULE_orderingComparator = 8
    RULE_stringLiteral = 9
    RULE_startStopQualifier = 10
    RULE_withinQualifier = 11
    RULE_repeatedQualifier = 12
    RULE_objectPath = 13
    RULE_objectType = 14
    RULE_firstPathComponent = 15
    RULE_objectPathComponent = 16
    RULE_setLiteral = 17
    RULE_primitiveLiteral = 18
    RULE_orderableLiteral = 19

    ruleNames =  [ "pattern", "observationExpressions", "observationExpressionOr", 
                   "observationExpressionAnd", "observationExpression", 
                   "comparisonExpression", "comparisonExpressionAnd", "propTest", 
                   "orderingComparator", "stringLiteral", "startStopQualifier", 
                   "withinQualifier", "repeatedQualifier", "objectPath", 
                   "objectType", "firstPathComponent", "objectPathComponent", 
                   "setLiteral", "primitiveLiteral", "orderableLiteral" ]

    EOF = Token.EOF
    IntNegLiteral=1
    IntPosLiteral=2
    FloatNegLiteral=3
    FloatPosLiteral=4
    HexLiteral=5
    BinaryLiteral=6
    StringLiteral=7
    BoolLiteral=8
    TimestampLiteral=9
    AND=10
    OR=11
    NOT=12
    FOLLOWEDBY=13
    LIKE=14
    MATCHES=15
    ISSUPERSET=16
    ISSUBSET=17
    LAST=18
    IN=19
    START=20
    STOP=21
    SECONDS=22
    TRUE=23
    FALSE=24
    WITHIN=25
    REPEATS=26
    TIMES=27
    IdentifierWithoutHyphen=28
    IdentifierWithHyphen=29
    EQ=30
    NEQ=31
    LT=32
    LE=33
    GT=34
    GE=35
    QUOTE=36
    COLON=37
    DOT=38
    COMMA=39
    RPAREN=40
    LPAREN=41
    RBRACK=42
    LBRACK=43
    PLUS=44
    HYPHEN=45
    MINUS=46
    POWER_OP=47
    DIVIDE=48
    ASTERISK=49
    WS=50
    COMMENT=51
    LINE_COMMENT=52
    InvalidCharacter=53

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.1")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class PatternContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def observationExpressions(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionsContext,0)


        def EOF(self):
            return self.getToken(STIXPatternParser.EOF, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_pattern

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPattern" ):
                listener.enterPattern(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPattern" ):
                listener.exitPattern(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPattern" ):
                return visitor.visitPattern(self)
            else:
                return visitor.visitChildren(self)




    def pattern(self):

        localctx = STIXPatternParser.PatternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.observationExpressions(0)
            self.state = 41
            self.match(STIXPatternParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObservationExpressionsContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def observationExpressionOr(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionOrContext,0)


        def observationExpressions(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ObservationExpressionsContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionsContext,i)


        def FOLLOWEDBY(self):
            return self.getToken(STIXPatternParser.FOLLOWEDBY, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_observationExpressions

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressions" ):
                listener.enterObservationExpressions(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressions" ):
                listener.exitObservationExpressions(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressions" ):
                return visitor.visitObservationExpressions(self)
            else:
                return visitor.visitChildren(self)



    def observationExpressions(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ObservationExpressionsContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 2
        self.enterRecursionRule(localctx, 2, self.RULE_observationExpressions, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 44
            self.observationExpressionOr(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 51
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,0,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.ObservationExpressionsContext(self, _parentctx, _parentState)
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpressions)
                    self.state = 46
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 47
                    self.match(STIXPatternParser.FOLLOWEDBY)
                    self.state = 48
                    self.observationExpressions(3) 
                self.state = 53
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,0,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ObservationExpressionOrContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def observationExpressionAnd(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionAndContext,0)


        def observationExpressionOr(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ObservationExpressionOrContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionOrContext,i)


        def OR(self):
            return self.getToken(STIXPatternParser.OR, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_observationExpressionOr

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionOr" ):
                listener.enterObservationExpressionOr(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionOr" ):
                listener.exitObservationExpressionOr(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionOr" ):
                return visitor.visitObservationExpressionOr(self)
            else:
                return visitor.visitChildren(self)



    def observationExpressionOr(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ObservationExpressionOrContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 4
        self.enterRecursionRule(localctx, 4, self.RULE_observationExpressionOr, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 55
            self.observationExpressionAnd(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 62
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,1,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.ObservationExpressionOrContext(self, _parentctx, _parentState)
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpressionOr)
                    self.state = 57
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 58
                    self.match(STIXPatternParser.OR)
                    self.state = 59
                    self.observationExpressionOr(3) 
                self.state = 64
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ObservationExpressionAndContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def observationExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionContext,0)


        def observationExpressionAnd(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ObservationExpressionAndContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionAndContext,i)


        def AND(self):
            return self.getToken(STIXPatternParser.AND, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_observationExpressionAnd

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionAnd" ):
                listener.enterObservationExpressionAnd(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionAnd" ):
                listener.exitObservationExpressionAnd(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionAnd" ):
                return visitor.visitObservationExpressionAnd(self)
            else:
                return visitor.visitChildren(self)



    def observationExpressionAnd(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ObservationExpressionAndContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 6
        self.enterRecursionRule(localctx, 6, self.RULE_observationExpressionAnd, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self.observationExpression(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 73
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,2,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.ObservationExpressionAndContext(self, _parentctx, _parentState)
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpressionAnd)
                    self.state = 68
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 69
                    self.match(STIXPatternParser.AND)
                    self.state = 70
                    self.observationExpressionAnd(3) 
                self.state = 75
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,2,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ObservationExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return STIXPatternParser.RULE_observationExpression

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class ObservationExpressionRepeatedContext(ObservationExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObservationExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def observationExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionContext,0)

        def repeatedQualifier(self):
            return self.getTypedRuleContext(STIXPatternParser.RepeatedQualifierContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionRepeated" ):
                listener.enterObservationExpressionRepeated(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionRepeated" ):
                listener.exitObservationExpressionRepeated(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionRepeated" ):
                return visitor.visitObservationExpressionRepeated(self)
            else:
                return visitor.visitChildren(self)


    class ObservationExpressionSimpleContext(ObservationExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObservationExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACK(self):
            return self.getToken(STIXPatternParser.LBRACK, 0)
        def comparisonExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ComparisonExpressionContext,0)

        def RBRACK(self):
            return self.getToken(STIXPatternParser.RBRACK, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionSimple" ):
                listener.enterObservationExpressionSimple(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionSimple" ):
                listener.exitObservationExpressionSimple(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionSimple" ):
                return visitor.visitObservationExpressionSimple(self)
            else:
                return visitor.visitChildren(self)


    class ObservationExpressionCompoundContext(ObservationExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObservationExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LPAREN(self):
            return self.getToken(STIXPatternParser.LPAREN, 0)
        def observationExpressions(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionsContext,0)

        def RPAREN(self):
            return self.getToken(STIXPatternParser.RPAREN, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionCompound" ):
                listener.enterObservationExpressionCompound(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionCompound" ):
                listener.exitObservationExpressionCompound(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionCompound" ):
                return visitor.visitObservationExpressionCompound(self)
            else:
                return visitor.visitChildren(self)


    class ObservationExpressionWithinContext(ObservationExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObservationExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def observationExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionContext,0)

        def withinQualifier(self):
            return self.getTypedRuleContext(STIXPatternParser.WithinQualifierContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionWithin" ):
                listener.enterObservationExpressionWithin(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionWithin" ):
                listener.exitObservationExpressionWithin(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionWithin" ):
                return visitor.visitObservationExpressionWithin(self)
            else:
                return visitor.visitChildren(self)


    class ObservationExpressionStartStopContext(ObservationExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObservationExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def observationExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ObservationExpressionContext,0)

        def startStopQualifier(self):
            return self.getTypedRuleContext(STIXPatternParser.StartStopQualifierContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObservationExpressionStartStop" ):
                listener.enterObservationExpressionStartStop(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObservationExpressionStartStop" ):
                listener.exitObservationExpressionStartStop(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObservationExpressionStartStop" ):
                return visitor.visitObservationExpressionStartStop(self)
            else:
                return visitor.visitChildren(self)



    def observationExpression(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ObservationExpressionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 8
        self.enterRecursionRule(localctx, 8, self.RULE_observationExpression, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 85
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [STIXPatternParser.LBRACK]:
                localctx = STIXPatternParser.ObservationExpressionSimpleContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 77
                self.match(STIXPatternParser.LBRACK)
                self.state = 78
                self.comparisonExpression(0)
                self.state = 79
                self.match(STIXPatternParser.RBRACK)
                pass
            elif token in [STIXPatternParser.LPAREN]:
                localctx = STIXPatternParser.ObservationExpressionCompoundContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 81
                self.match(STIXPatternParser.LPAREN)
                self.state = 82
                self.observationExpressions(0)
                self.state = 83
                self.match(STIXPatternParser.RPAREN)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 95
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,5,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 93
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,4,self._ctx)
                    if la_ == 1:
                        localctx = STIXPatternParser.ObservationExpressionStartStopContext(self, STIXPatternParser.ObservationExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpression)
                        self.state = 87
                        if not self.precpred(self._ctx, 3):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                        self.state = 88
                        self.startStopQualifier()
                        pass

                    elif la_ == 2:
                        localctx = STIXPatternParser.ObservationExpressionWithinContext(self, STIXPatternParser.ObservationExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpression)
                        self.state = 89
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 90
                        self.withinQualifier()
                        pass

                    elif la_ == 3:
                        localctx = STIXPatternParser.ObservationExpressionRepeatedContext(self, STIXPatternParser.ObservationExpressionContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_observationExpression)
                        self.state = 91
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 92
                        self.repeatedQualifier()
                        pass

             
                self.state = 97
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,5,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ComparisonExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return STIXPatternParser.RULE_comparisonExpression

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class ComparisonExpressionAnd_Context(ComparisonExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ComparisonExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def comparisonExpressionAnd(self):
            return self.getTypedRuleContext(STIXPatternParser.ComparisonExpressionAndContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparisonExpressionAnd_" ):
                listener.enterComparisonExpressionAnd_(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparisonExpressionAnd_" ):
                listener.exitComparisonExpressionAnd_(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComparisonExpressionAnd_" ):
                return visitor.visitComparisonExpressionAnd_(self)
            else:
                return visitor.visitChildren(self)


    class ComparisonExpressionOredContext(ComparisonExpressionContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ComparisonExpressionContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def comparisonExpression(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ComparisonExpressionContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ComparisonExpressionContext,i)

        def OR(self):
            return self.getToken(STIXPatternParser.OR, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparisonExpressionOred" ):
                listener.enterComparisonExpressionOred(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparisonExpressionOred" ):
                listener.exitComparisonExpressionOred(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComparisonExpressionOred" ):
                return visitor.visitComparisonExpressionOred(self)
            else:
                return visitor.visitChildren(self)



    def comparisonExpression(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ComparisonExpressionContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 10
        self.enterRecursionRule(localctx, 10, self.RULE_comparisonExpression, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            localctx = STIXPatternParser.ComparisonExpressionAnd_Context(self, localctx)
            self._ctx = localctx
            _prevctx = localctx

            self.state = 99
            self.comparisonExpressionAnd(0)
            self._ctx.stop = self._input.LT(-1)
            self.state = 106
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,6,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.ComparisonExpressionOredContext(self, STIXPatternParser.ComparisonExpressionContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_comparisonExpression)
                    self.state = 101
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 102
                    self.match(STIXPatternParser.OR)
                    self.state = 103
                    self.comparisonExpression(3) 
                self.state = 108
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,6,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class ComparisonExpressionAndContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return STIXPatternParser.RULE_comparisonExpressionAnd

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class ComparisonExpressionAndPropTestContext(ComparisonExpressionAndContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ComparisonExpressionAndContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def propTest(self):
            return self.getTypedRuleContext(STIXPatternParser.PropTestContext,0)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparisonExpressionAndPropTest" ):
                listener.enterComparisonExpressionAndPropTest(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparisonExpressionAndPropTest" ):
                listener.exitComparisonExpressionAndPropTest(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComparisonExpressionAndPropTest" ):
                return visitor.visitComparisonExpressionAndPropTest(self)
            else:
                return visitor.visitChildren(self)


    class ComparisonExpressionAndedContext(ComparisonExpressionAndContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ComparisonExpressionAndContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def comparisonExpressionAnd(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ComparisonExpressionAndContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ComparisonExpressionAndContext,i)

        def AND(self):
            return self.getToken(STIXPatternParser.AND, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterComparisonExpressionAnded" ):
                listener.enterComparisonExpressionAnded(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitComparisonExpressionAnded" ):
                listener.exitComparisonExpressionAnded(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitComparisonExpressionAnded" ):
                return visitor.visitComparisonExpressionAnded(self)
            else:
                return visitor.visitChildren(self)



    def comparisonExpressionAnd(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ComparisonExpressionAndContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 12
        self.enterRecursionRule(localctx, 12, self.RULE_comparisonExpressionAnd, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            localctx = STIXPatternParser.ComparisonExpressionAndPropTestContext(self, localctx)
            self._ctx = localctx
            _prevctx = localctx

            self.state = 110
            self.propTest()
            self._ctx.stop = self._input.LT(-1)
            self.state = 117
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.ComparisonExpressionAndedContext(self, STIXPatternParser.ComparisonExpressionAndContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_comparisonExpressionAnd)
                    self.state = 112
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 113
                    self.match(STIXPatternParser.AND)
                    self.state = 114
                    self.comparisonExpressionAnd(3) 
                self.state = 119
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class PropTestContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return STIXPatternParser.RULE_propTest

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class PropTestRegexContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def MATCHES(self):
            return self.getToken(STIXPatternParser.MATCHES, 0)
        def stringLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.StringLiteralContext,0)

        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestRegex" ):
                listener.enterPropTestRegex(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestRegex" ):
                listener.exitPropTestRegex(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestRegex" ):
                return visitor.visitPropTestRegex(self)
            else:
                return visitor.visitChildren(self)


    class PropTestOrderContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def orderableLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.OrderableLiteralContext,0)

        def GT(self):
            return self.getToken(STIXPatternParser.GT, 0)
        def LT(self):
            return self.getToken(STIXPatternParser.LT, 0)
        def GE(self):
            return self.getToken(STIXPatternParser.GE, 0)
        def LE(self):
            return self.getToken(STIXPatternParser.LE, 0)
        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestOrder" ):
                listener.enterPropTestOrder(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestOrder" ):
                listener.exitPropTestOrder(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestOrder" ):
                return visitor.visitPropTestOrder(self)
            else:
                return visitor.visitChildren(self)


    class PropTestLikeContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def LIKE(self):
            return self.getToken(STIXPatternParser.LIKE, 0)
        def stringLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.StringLiteralContext,0)

        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestLike" ):
                listener.enterPropTestLike(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestLike" ):
                listener.exitPropTestLike(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestLike" ):
                return visitor.visitPropTestLike(self)
            else:
                return visitor.visitChildren(self)


    class PropTestEqualContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def primitiveLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.PrimitiveLiteralContext,0)

        def EQ(self):
            return self.getToken(STIXPatternParser.EQ, 0)
        def NEQ(self):
            return self.getToken(STIXPatternParser.NEQ, 0)
        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestEqual" ):
                listener.enterPropTestEqual(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestEqual" ):
                listener.exitPropTestEqual(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestEqual" ):
                return visitor.visitPropTestEqual(self)
            else:
                return visitor.visitChildren(self)


    class PropTestSetContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def IN(self):
            return self.getToken(STIXPatternParser.IN, 0)
        def setLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.SetLiteralContext,0)

        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestSet" ):
                listener.enterPropTestSet(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestSet" ):
                listener.exitPropTestSet(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestSet" ):
                return visitor.visitPropTestSet(self)
            else:
                return visitor.visitChildren(self)


    class PropTestIsSubsetContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def ISSUBSET(self):
            return self.getToken(STIXPatternParser.ISSUBSET, 0)
        def stringLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.StringLiteralContext,0)

        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestIsSubset" ):
                listener.enterPropTestIsSubset(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestIsSubset" ):
                listener.exitPropTestIsSubset(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestIsSubset" ):
                return visitor.visitPropTestIsSubset(self)
            else:
                return visitor.visitChildren(self)


    class PropTestParenContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LPAREN(self):
            return self.getToken(STIXPatternParser.LPAREN, 0)
        def comparisonExpression(self):
            return self.getTypedRuleContext(STIXPatternParser.ComparisonExpressionContext,0)

        def RPAREN(self):
            return self.getToken(STIXPatternParser.RPAREN, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestParen" ):
                listener.enterPropTestParen(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestParen" ):
                listener.exitPropTestParen(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestParen" ):
                return visitor.visitPropTestParen(self)
            else:
                return visitor.visitChildren(self)


    class PropTestIsSupersetContext(PropTestContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.PropTestContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPath(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathContext,0)

        def ISSUPERSET(self):
            return self.getToken(STIXPatternParser.ISSUPERSET, 0)
        def stringLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.StringLiteralContext,0)

        def NOT(self):
            return self.getToken(STIXPatternParser.NOT, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropTestIsSuperset" ):
                listener.enterPropTestIsSuperset(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropTestIsSuperset" ):
                listener.exitPropTestIsSuperset(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPropTestIsSuperset" ):
                return visitor.visitPropTestIsSuperset(self)
            else:
                return visitor.visitChildren(self)



    def propTest(self):

        localctx = STIXPatternParser.PropTestContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_propTest)
        self._la = 0 # Token type
        try:
            self.state = 173
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,15,self._ctx)
            if la_ == 1:
                localctx = STIXPatternParser.PropTestEqualContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 120
                self.objectPath()
                self.state = 122
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 121
                    self.match(STIXPatternParser.NOT)


                self.state = 124
                _la = self._input.LA(1)
                if not(_la==STIXPatternParser.EQ or _la==STIXPatternParser.NEQ):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 125
                self.primitiveLiteral()
                pass

            elif la_ == 2:
                localctx = STIXPatternParser.PropTestOrderContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 127
                self.objectPath()
                self.state = 129
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 128
                    self.match(STIXPatternParser.NOT)


                self.state = 131
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << STIXPatternParser.LT) | (1 << STIXPatternParser.LE) | (1 << STIXPatternParser.GT) | (1 << STIXPatternParser.GE))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 132
                self.orderableLiteral()
                pass

            elif la_ == 3:
                localctx = STIXPatternParser.PropTestSetContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 134
                self.objectPath()
                self.state = 136
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 135
                    self.match(STIXPatternParser.NOT)


                self.state = 138
                self.match(STIXPatternParser.IN)
                self.state = 139
                self.setLiteral()
                pass

            elif la_ == 4:
                localctx = STIXPatternParser.PropTestLikeContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 141
                self.objectPath()
                self.state = 143
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 142
                    self.match(STIXPatternParser.NOT)


                self.state = 145
                self.match(STIXPatternParser.LIKE)
                self.state = 146
                self.stringLiteral()
                pass

            elif la_ == 5:
                localctx = STIXPatternParser.PropTestRegexContext(self, localctx)
                self.enterOuterAlt(localctx, 5)
                self.state = 148
                self.objectPath()
                self.state = 150
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 149
                    self.match(STIXPatternParser.NOT)


                self.state = 152
                self.match(STIXPatternParser.MATCHES)
                self.state = 153
                self.stringLiteral()
                pass

            elif la_ == 6:
                localctx = STIXPatternParser.PropTestIsSubsetContext(self, localctx)
                self.enterOuterAlt(localctx, 6)
                self.state = 155
                self.objectPath()
                self.state = 157
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 156
                    self.match(STIXPatternParser.NOT)


                self.state = 159
                self.match(STIXPatternParser.ISSUBSET)
                self.state = 160
                self.stringLiteral()
                pass

            elif la_ == 7:
                localctx = STIXPatternParser.PropTestIsSupersetContext(self, localctx)
                self.enterOuterAlt(localctx, 7)
                self.state = 162
                self.objectPath()
                self.state = 164
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==STIXPatternParser.NOT:
                    self.state = 163
                    self.match(STIXPatternParser.NOT)


                self.state = 166
                self.match(STIXPatternParser.ISSUPERSET)
                self.state = 167
                self.stringLiteral()
                pass

            elif la_ == 8:
                localctx = STIXPatternParser.PropTestParenContext(self, localctx)
                self.enterOuterAlt(localctx, 8)
                self.state = 169
                self.match(STIXPatternParser.LPAREN)
                self.state = 170
                self.comparisonExpression(0)
                self.state = 171
                self.match(STIXPatternParser.RPAREN)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OrderingComparatorContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def GT(self):
            return self.getToken(STIXPatternParser.GT, 0)

        def LT(self):
            return self.getToken(STIXPatternParser.LT, 0)

        def GE(self):
            return self.getToken(STIXPatternParser.GE, 0)

        def LE(self):
            return self.getToken(STIXPatternParser.LE, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_orderingComparator

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOrderingComparator" ):
                listener.enterOrderingComparator(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOrderingComparator" ):
                listener.exitOrderingComparator(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOrderingComparator" ):
                return visitor.visitOrderingComparator(self)
            else:
                return visitor.visitChildren(self)




    def orderingComparator(self):

        localctx = STIXPatternParser.OrderingComparatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_orderingComparator)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 175
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << STIXPatternParser.LT) | (1 << STIXPatternParser.LE) | (1 << STIXPatternParser.GT) | (1 << STIXPatternParser.GE))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StringLiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def StringLiteral(self):
            return self.getToken(STIXPatternParser.StringLiteral, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_stringLiteral

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStringLiteral" ):
                listener.enterStringLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStringLiteral" ):
                listener.exitStringLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStringLiteral" ):
                return visitor.visitStringLiteral(self)
            else:
                return visitor.visitChildren(self)




    def stringLiteral(self):

        localctx = STIXPatternParser.StringLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_stringLiteral)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 177
            self.match(STIXPatternParser.StringLiteral)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class StartStopQualifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def START(self):
            return self.getToken(STIXPatternParser.START, 0)

        def TimestampLiteral(self, i:int=None):
            if i is None:
                return self.getTokens(STIXPatternParser.TimestampLiteral)
            else:
                return self.getToken(STIXPatternParser.TimestampLiteral, i)

        def STOP(self):
            return self.getToken(STIXPatternParser.STOP, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_startStopQualifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterStartStopQualifier" ):
                listener.enterStartStopQualifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitStartStopQualifier" ):
                listener.exitStartStopQualifier(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitStartStopQualifier" ):
                return visitor.visitStartStopQualifier(self)
            else:
                return visitor.visitChildren(self)




    def startStopQualifier(self):

        localctx = STIXPatternParser.StartStopQualifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_startStopQualifier)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 179
            self.match(STIXPatternParser.START)
            self.state = 180
            self.match(STIXPatternParser.TimestampLiteral)
            self.state = 181
            self.match(STIXPatternParser.STOP)
            self.state = 182
            self.match(STIXPatternParser.TimestampLiteral)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class WithinQualifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def WITHIN(self):
            return self.getToken(STIXPatternParser.WITHIN, 0)

        def SECONDS(self):
            return self.getToken(STIXPatternParser.SECONDS, 0)

        def IntPosLiteral(self):
            return self.getToken(STIXPatternParser.IntPosLiteral, 0)

        def FloatPosLiteral(self):
            return self.getToken(STIXPatternParser.FloatPosLiteral, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_withinQualifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterWithinQualifier" ):
                listener.enterWithinQualifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitWithinQualifier" ):
                listener.exitWithinQualifier(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitWithinQualifier" ):
                return visitor.visitWithinQualifier(self)
            else:
                return visitor.visitChildren(self)




    def withinQualifier(self):

        localctx = STIXPatternParser.WithinQualifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_withinQualifier)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 184
            self.match(STIXPatternParser.WITHIN)
            self.state = 185
            _la = self._input.LA(1)
            if not(_la==STIXPatternParser.IntPosLiteral or _la==STIXPatternParser.FloatPosLiteral):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 186
            self.match(STIXPatternParser.SECONDS)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RepeatedQualifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def REPEATS(self):
            return self.getToken(STIXPatternParser.REPEATS, 0)

        def IntPosLiteral(self):
            return self.getToken(STIXPatternParser.IntPosLiteral, 0)

        def TIMES(self):
            return self.getToken(STIXPatternParser.TIMES, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_repeatedQualifier

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterRepeatedQualifier" ):
                listener.enterRepeatedQualifier(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitRepeatedQualifier" ):
                listener.exitRepeatedQualifier(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRepeatedQualifier" ):
                return visitor.visitRepeatedQualifier(self)
            else:
                return visitor.visitChildren(self)




    def repeatedQualifier(self):

        localctx = STIXPatternParser.RepeatedQualifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_repeatedQualifier)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 188
            self.match(STIXPatternParser.REPEATS)
            self.state = 189
            self.match(STIXPatternParser.IntPosLiteral)
            self.state = 190
            self.match(STIXPatternParser.TIMES)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectPathContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def objectType(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectTypeContext,0)


        def COLON(self):
            return self.getToken(STIXPatternParser.COLON, 0)

        def firstPathComponent(self):
            return self.getTypedRuleContext(STIXPatternParser.FirstPathComponentContext,0)


        def objectPathComponent(self):
            return self.getTypedRuleContext(STIXPatternParser.ObjectPathComponentContext,0)


        def getRuleIndex(self):
            return STIXPatternParser.RULE_objectPath

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObjectPath" ):
                listener.enterObjectPath(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObjectPath" ):
                listener.exitObjectPath(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectPath" ):
                return visitor.visitObjectPath(self)
            else:
                return visitor.visitChildren(self)




    def objectPath(self):

        localctx = STIXPatternParser.ObjectPathContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_objectPath)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 192
            self.objectType()
            self.state = 193
            self.match(STIXPatternParser.COLON)
            self.state = 194
            self.firstPathComponent()
            self.state = 196
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==STIXPatternParser.DOT or _la==STIXPatternParser.LBRACK:
                self.state = 195
                self.objectPathComponent(0)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectTypeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IdentifierWithoutHyphen(self):
            return self.getToken(STIXPatternParser.IdentifierWithoutHyphen, 0)

        def IdentifierWithHyphen(self):
            return self.getToken(STIXPatternParser.IdentifierWithHyphen, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_objectType

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObjectType" ):
                listener.enterObjectType(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObjectType" ):
                listener.exitObjectType(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectType" ):
                return visitor.visitObjectType(self)
            else:
                return visitor.visitChildren(self)




    def objectType(self):

        localctx = STIXPatternParser.ObjectTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_objectType)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 198
            _la = self._input.LA(1)
            if not(_la==STIXPatternParser.IdentifierWithoutHyphen or _la==STIXPatternParser.IdentifierWithHyphen):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FirstPathComponentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IdentifierWithoutHyphen(self):
            return self.getToken(STIXPatternParser.IdentifierWithoutHyphen, 0)

        def StringLiteral(self):
            return self.getToken(STIXPatternParser.StringLiteral, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_firstPathComponent

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterFirstPathComponent" ):
                listener.enterFirstPathComponent(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitFirstPathComponent" ):
                listener.exitFirstPathComponent(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitFirstPathComponent" ):
                return visitor.visitFirstPathComponent(self)
            else:
                return visitor.visitChildren(self)




    def firstPathComponent(self):

        localctx = STIXPatternParser.FirstPathComponentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_firstPathComponent)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 200
            _la = self._input.LA(1)
            if not(_la==STIXPatternParser.StringLiteral or _la==STIXPatternParser.IdentifierWithoutHyphen):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectPathComponentContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return STIXPatternParser.RULE_objectPathComponent

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)


    class IndexPathStepContext(ObjectPathComponentContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObjectPathComponentContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACK(self):
            return self.getToken(STIXPatternParser.LBRACK, 0)
        def RBRACK(self):
            return self.getToken(STIXPatternParser.RBRACK, 0)
        def IntPosLiteral(self):
            return self.getToken(STIXPatternParser.IntPosLiteral, 0)
        def IntNegLiteral(self):
            return self.getToken(STIXPatternParser.IntNegLiteral, 0)
        def ASTERISK(self):
            return self.getToken(STIXPatternParser.ASTERISK, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterIndexPathStep" ):
                listener.enterIndexPathStep(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitIndexPathStep" ):
                listener.exitIndexPathStep(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIndexPathStep" ):
                return visitor.visitIndexPathStep(self)
            else:
                return visitor.visitChildren(self)


    class PathStepContext(ObjectPathComponentContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObjectPathComponentContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def objectPathComponent(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.ObjectPathComponentContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.ObjectPathComponentContext,i)


        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPathStep" ):
                listener.enterPathStep(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPathStep" ):
                listener.exitPathStep(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPathStep" ):
                return visitor.visitPathStep(self)
            else:
                return visitor.visitChildren(self)


    class KeyPathStepContext(ObjectPathComponentContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a STIXPatternParser.ObjectPathComponentContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def IdentifierWithoutHyphen(self):
            return self.getToken(STIXPatternParser.IdentifierWithoutHyphen, 0)
        def StringLiteral(self):
            return self.getToken(STIXPatternParser.StringLiteral, 0)

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterKeyPathStep" ):
                listener.enterKeyPathStep(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitKeyPathStep" ):
                listener.exitKeyPathStep(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitKeyPathStep" ):
                return visitor.visitKeyPathStep(self)
            else:
                return visitor.visitChildren(self)



    def objectPathComponent(self, _p:int=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = STIXPatternParser.ObjectPathComponentContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 32
        self.enterRecursionRule(localctx, 32, self.RULE_objectPathComponent, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 208
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [STIXPatternParser.DOT]:
                localctx = STIXPatternParser.KeyPathStepContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 203
                self.match(STIXPatternParser.DOT)
                self.state = 204
                _la = self._input.LA(1)
                if not(_la==STIXPatternParser.StringLiteral or _la==STIXPatternParser.IdentifierWithoutHyphen):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                pass
            elif token in [STIXPatternParser.LBRACK]:
                localctx = STIXPatternParser.IndexPathStepContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 205
                self.match(STIXPatternParser.LBRACK)
                self.state = 206
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << STIXPatternParser.IntNegLiteral) | (1 << STIXPatternParser.IntPosLiteral) | (1 << STIXPatternParser.ASTERISK))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 207
                self.match(STIXPatternParser.RBRACK)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 214
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,18,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = STIXPatternParser.PathStepContext(self, STIXPatternParser.ObjectPathComponentContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_objectPathComponent)
                    self.state = 210
                    if not self.precpred(self._ctx, 3):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 3)")
                    self.state = 211
                    self.objectPathComponent(4) 
                self.state = 216
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,18,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class SetLiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LPAREN(self):
            return self.getToken(STIXPatternParser.LPAREN, 0)

        def RPAREN(self):
            return self.getToken(STIXPatternParser.RPAREN, 0)

        def primitiveLiteral(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(STIXPatternParser.PrimitiveLiteralContext)
            else:
                return self.getTypedRuleContext(STIXPatternParser.PrimitiveLiteralContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(STIXPatternParser.COMMA)
            else:
                return self.getToken(STIXPatternParser.COMMA, i)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_setLiteral

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterSetLiteral" ):
                listener.enterSetLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitSetLiteral" ):
                listener.exitSetLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSetLiteral" ):
                return visitor.visitSetLiteral(self)
            else:
                return visitor.visitChildren(self)




    def setLiteral(self):

        localctx = STIXPatternParser.SetLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_setLiteral)
        self._la = 0 # Token type
        try:
            self.state = 230
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,20,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 217
                self.match(STIXPatternParser.LPAREN)
                self.state = 218
                self.match(STIXPatternParser.RPAREN)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 219
                self.match(STIXPatternParser.LPAREN)
                self.state = 220
                self.primitiveLiteral()
                self.state = 225
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==STIXPatternParser.COMMA:
                    self.state = 221
                    self.match(STIXPatternParser.COMMA)
                    self.state = 222
                    self.primitiveLiteral()
                    self.state = 227
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 228
                self.match(STIXPatternParser.RPAREN)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PrimitiveLiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def orderableLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.OrderableLiteralContext,0)


        def BoolLiteral(self):
            return self.getToken(STIXPatternParser.BoolLiteral, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_primitiveLiteral

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPrimitiveLiteral" ):
                listener.enterPrimitiveLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPrimitiveLiteral" ):
                listener.exitPrimitiveLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPrimitiveLiteral" ):
                return visitor.visitPrimitiveLiteral(self)
            else:
                return visitor.visitChildren(self)




    def primitiveLiteral(self):

        localctx = STIXPatternParser.PrimitiveLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_primitiveLiteral)
        try:
            self.state = 234
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [STIXPatternParser.IntNegLiteral, STIXPatternParser.IntPosLiteral, STIXPatternParser.FloatNegLiteral, STIXPatternParser.FloatPosLiteral, STIXPatternParser.HexLiteral, STIXPatternParser.BinaryLiteral, STIXPatternParser.StringLiteral, STIXPatternParser.TimestampLiteral]:
                self.enterOuterAlt(localctx, 1)
                self.state = 232
                self.orderableLiteral()
                pass
            elif token in [STIXPatternParser.BoolLiteral]:
                self.enterOuterAlt(localctx, 2)
                self.state = 233
                self.match(STIXPatternParser.BoolLiteral)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OrderableLiteralContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IntPosLiteral(self):
            return self.getToken(STIXPatternParser.IntPosLiteral, 0)

        def IntNegLiteral(self):
            return self.getToken(STIXPatternParser.IntNegLiteral, 0)

        def FloatPosLiteral(self):
            return self.getToken(STIXPatternParser.FloatPosLiteral, 0)

        def FloatNegLiteral(self):
            return self.getToken(STIXPatternParser.FloatNegLiteral, 0)

        def stringLiteral(self):
            return self.getTypedRuleContext(STIXPatternParser.StringLiteralContext,0)


        def BinaryLiteral(self):
            return self.getToken(STIXPatternParser.BinaryLiteral, 0)

        def HexLiteral(self):
            return self.getToken(STIXPatternParser.HexLiteral, 0)

        def TimestampLiteral(self):
            return self.getToken(STIXPatternParser.TimestampLiteral, 0)

        def getRuleIndex(self):
            return STIXPatternParser.RULE_orderableLiteral

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterOrderableLiteral" ):
                listener.enterOrderableLiteral(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitOrderableLiteral" ):
                listener.exitOrderableLiteral(self)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOrderableLiteral" ):
                return visitor.visitOrderableLiteral(self)
            else:
                return visitor.visitChildren(self)




    def orderableLiteral(self):

        localctx = STIXPatternParser.OrderableLiteralContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_orderableLiteral)
        try:
            self.state = 244
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [STIXPatternParser.IntPosLiteral]:
                self.enterOuterAlt(localctx, 1)
                self.state = 236
                self.match(STIXPatternParser.IntPosLiteral)
                pass
            elif token in [STIXPatternParser.IntNegLiteral]:
                self.enterOuterAlt(localctx, 2)
                self.state = 237
                self.match(STIXPatternParser.IntNegLiteral)
                pass
            elif token in [STIXPatternParser.FloatPosLiteral]:
                self.enterOuterAlt(localctx, 3)
                self.state = 238
                self.match(STIXPatternParser.FloatPosLiteral)
                pass
            elif token in [STIXPatternParser.FloatNegLiteral]:
                self.enterOuterAlt(localctx, 4)
                self.state = 239
                self.match(STIXPatternParser.FloatNegLiteral)
                pass
            elif token in [STIXPatternParser.StringLiteral]:
                self.enterOuterAlt(localctx, 5)
                self.state = 240
                self.stringLiteral()
                pass
            elif token in [STIXPatternParser.BinaryLiteral]:
                self.enterOuterAlt(localctx, 6)
                self.state = 241
                self.match(STIXPatternParser.BinaryLiteral)
                pass
            elif token in [STIXPatternParser.HexLiteral]:
                self.enterOuterAlt(localctx, 7)
                self.state = 242
                self.match(STIXPatternParser.HexLiteral)
                pass
            elif token in [STIXPatternParser.TimestampLiteral]:
                self.enterOuterAlt(localctx, 8)
                self.state = 243
                self.match(STIXPatternParser.TimestampLiteral)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



    def sempred(self, localctx:RuleContext, ruleIndex:int, predIndex:int):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[1] = self.observationExpressions_sempred
        self._predicates[2] = self.observationExpressionOr_sempred
        self._predicates[3] = self.observationExpressionAnd_sempred
        self._predicates[4] = self.observationExpression_sempred
        self._predicates[5] = self.comparisonExpression_sempred
        self._predicates[6] = self.comparisonExpressionAnd_sempred
        self._predicates[16] = self.objectPathComponent_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def observationExpressions_sempred(self, localctx:ObservationExpressionsContext, predIndex:int):
            if predIndex == 0:
                return self.precpred(self._ctx, 2)
         

    def observationExpressionOr_sempred(self, localctx:ObservationExpressionOrContext, predIndex:int):
            if predIndex == 1:
                return self.precpred(self._ctx, 2)
         

    def observationExpressionAnd_sempred(self, localctx:ObservationExpressionAndContext, predIndex:int):
            if predIndex == 2:
                return self.precpred(self._ctx, 2)
         

    def observationExpression_sempred(self, localctx:ObservationExpressionContext, predIndex:int):
            if predIndex == 3:
                return self.precpred(self._ctx, 3)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 5:
                return self.precpred(self._ctx, 1)
         

    def comparisonExpression_sempred(self, localctx:ComparisonExpressionContext, predIndex:int):
            if predIndex == 6:
                return self.precpred(self._ctx, 2)
         

    def comparisonExpressionAnd_sempred(self, localctx:ComparisonExpressionAndContext, predIndex:int):
            if predIndex == 7:
                return self.precpred(self._ctx, 2)
         

    def objectPathComponent_sempred(self, localctx:ObjectPathComponentContext, predIndex:int):
            if predIndex == 8:
                return self.precpred(self._ctx, 3)
         







import json
from . import json_to_stix_translator
from stix_shifter_utils.modules.base.stix_translation.base_results_translator import BaseResultTranslator
from stix_shifter_utils.stix_translation.src.utils import transformers
from stix_shifter_utils.stix_translation.src.utils.exceptions import LoadJsonResultsException, TranslationResultException

# Concrete BaseResultTranslator


class JSONToStix(BaseResultTranslator):

    def translate_results(self, data_source, data, options, mapping=None):
        """
        Translates JSON data into STIX results based on a mapping file
        :param data: JSON formatted data to translate into STIX format
        :type data: str
        :param mapping: The mapping file path to use as instructions on how to translate the given JSON data to STIX. 
            Defaults the path to whatever is passed into the constructor for JSONToSTIX (This should be the to_stix_map.json in the module's json directory)
        :type mapping: str (filepath)
        :return: STIX formatted results
        :rtype: str
        """
        self.mapping = options['mapping'] if 'mapping' in options else {}
        try:
            json_data = json.loads(data)
            data_source = json.loads(data_source)
        except Exception:
            raise LoadJsonResultsException()

        if(not self.mapping):
            map_file = open(self.default_mapping_file_path).read()
            map_data = json.loads(map_file)
        else:
            map_data = self.mapping

        try:
            results = json_to_stix_translator.convert_to_stix(data_source, map_data, json_data, transformers.get_all_transformers(), options, self.callback)
        except Exception as ex:
            raise TranslationResultException("Error when converting results to STIX: {}".format(ex))

        return json.dumps(results, indent=4, sort_keys=False)



from abc import ABCMeta, abstractmethod
from .base_results_translator import BaseResultTranslator
from .base_query_translator import BaseQueryTranslator


class BaseTranslator:

    def __init__(self):
        self.results_translator = BaseResultTranslator()
        self.query_translator = BaseQueryTranslator()

    def translate_results(self, data_source, data, options, mapping=None):
        """
        Translates data into STIX results based on a mapping file
        :param data_source: STIX identity object representing a data source
        :type data_source: str
        :param data: data to translate into STIX format
        :type data: str
        :param mapping: The mapping file path to use as instructions on how to translate the given data to STIX.
        :type mapping: str (filepath)
        :return: translated STIX formatted results
        :rtype: str
        """
        return self.results_translator.translate_results(data_source, data, options, mapping)

    def transform_query(self, data, antlr_parsing_object, data_model_mapper, options, mapping=None):
        """
        Transforms STIX query into a different query format. based on a mapping file
        :param data: STIX pattern to transform into native data source query
        :type data: str
        :param antlr_parsing_object: Antlr parsing objects for the STIX pattern
        :type antlr_parsing_object: object
        :data_model_mapper: Mapping object for the data source
        :type data_model_mapper: object
        :param mapping: The mapping file path to use as instructions on how to transform the given STIX pattern into native data source query.
        :type mapping: str (filepath)
        :return: transformed query string
        :rtype: str
        """

        return self.query_translator.transform_query(data, antlr_parsing_object, data_model_mapper, options, mapping)



#!/usr/bin/env python3
from datetime import datetime, timezone, tzinfo, timedelta
import base64
import socket
import re
import os
from urllib.parse import urlparse


class ValueTransformer():
    """ Base class for value transformers """

    @staticmethod
    def transform(obj):
        """ abstract function for converting value formats """
        raise NotImplementedError


class StringToBool(ValueTransformer):
    """A value transformer for converting String to boolean value"""

    @staticmethod
    def transform(value):
        return value.lower() in ("yes", "true", "t", "1")


class SplunkToTimestamp(ValueTransformer):
    """A value transformer for converting Splunk timestamp to regular timestamp"""

    @staticmethod
    def transform(splunkTime):
        return splunkTime[:-6]+'Z'


class EpochToTimestamp(ValueTransformer):
    """A value transformer for the timestamps"""

    @staticmethod
    def transform(epoch):
        try:
            return (datetime.fromtimestamp(int(epoch) / 1000, timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z')
        except ValueError:
            print("Cannot convert epoch value {} to timestamp".format(epoch))


class FormatMac(ValueTransformer):
    """A value transformer to convert Mac address to STIX Mac address format"""

    @staticmethod
    def transform(mac):
        value = ':'.join([mac[i:i + 2] for i in range(0, len(mac), 2)])
        return value.lower()


class MsatpToTimestamp(ValueTransformer):
    """A value transformer to truncate milliseconds"""

    @staticmethod
    def transform(msatptime):
        time_array = msatptime.split('.')
        converted_time = time_array[0] + '.' + time_array[1][:3] + 'Z' if len(time_array) > 1 else time_array[0] + 'Z'
        return converted_time


class FormatTCPProtocol(ValueTransformer):
    """A value transformer to convert TCP protocol to IANA format"""

    @staticmethod
    def transform(protocolname):
        converted_name = re.search(r'^tcp', protocolname, re.I).group(0)
        try:
            obj_array = converted_name if isinstance(converted_name, list) else converted_name.split(', ')
            # Loop through entries inside obj_array and make all strings lowercase to meet STIX format
            obj_array = [entry.lower() for entry in obj_array]
            return obj_array
        except ValueError:
            print("Cannot convert input to array")


class MsatpToRegistryValue(ValueTransformer):
    """A value transformer to convert MSATP Registry value protocol to windows-registry-value-type STIX"""

    @staticmethod
    def transform(registryvalues):
        stix_mapping = {"RegistryValueName": "name", "RegistryValueData": "data", "RegistryValueType": "data_type"}
        stix_datatype_mapping = {"None": "REG_NONE", "String": "REG_SZ", "Dword": "REG_DWORD",
                                 "ExpandString": "REG_EXPAND_SZ", "MultiString": "REG_MULTI_SZ",
                                 "Binary": "REG_BINARY", "Qword": "REG_QWORD"}
        converted_value = list()
        registryvalue_dict = dict()
        for each_value in registryvalues:
            for key, value in each_value.items():
                is_data_add = True
                if key == "RegistryValueType":
                    if value in stix_datatype_mapping.keys():
                        value = stix_datatype_mapping[value]
                    else:
                        is_data_add = False
                if is_data_add:
                    registryvalue_dict.update({stix_mapping[key]: value})
        converted_value.append(registryvalue_dict)
        return converted_value


class AwsToTimestamp(ValueTransformer):
    """
    A value transformer for converting AWS timestamp (YYYY-MM-DD hh:mm:ss.000)
    to UTC timestamp (YYYY-MM-DDThh:mm:ss.000Z)
    """

    @staticmethod
    def transform(aws_time):
        time_array = aws_time.split(' ')
        converted_time = time_array[0] + 'T' + time_array[1] + 'Z'
        return converted_time


class EpochSecondsToTimestamp(ValueTransformer):
    """A value transformer for the timestamps"""

    @staticmethod
    def transform(epoch):
        try:
            return (datetime.fromtimestamp(int(epoch), timezone.utc)
                    .strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z')
        except ValueError:
            print("Cannot convert epoch value {} to timestamp".format(epoch))


class TimestampToMilliseconds(ValueTransformer):
    """
    A value transformer for converting a UTC timestamp (YYYY-MM-DDThh:mm:ss.000Z) 
    to 13-digit Unix time (epoch + milliseconds)
    """

    @staticmethod
    def transform(timestamp):
        time_pattern = '%Y-%m-%dT%H:%M:%S.%fZ'
        epoch = datetime(1970, 1, 1)
        try:
            converted_time = int(((datetime.strptime(timestamp, time_pattern) - epoch).total_seconds()) * 1000)
            return converted_time
        except ValueError:
            print("Cannot convert the timestamp {} to milliseconds".format(timestamp))


class ToInteger(ValueTransformer):
    """A value transformer for expected integer values"""

    @staticmethod
    def transform(obj):
        try:
            if type(obj) is str and re.search('\.', obj):
                obj = float(obj)
            return int(obj)
        except ValueError:
            print("Cannot convert input {} to integer".format(obj))


class ToString(ValueTransformer):
    """A value transformer for expected string values"""

    @staticmethod
    def transform(obj):
        try:
            return str(obj)
        except ValueError:
            print("Cannot convert input to string")


class ToLowercaseArray(ValueTransformer):
    """A value transformer for expected array values"""

    @staticmethod
    def transform(obj):
        try:
            obj_array = obj if isinstance(obj, list) else obj.split(', ')
            # Loop through entries inside obj_array and make all strings lowercase to meet STIX format
            obj_array = [entry.lower() for entry in obj_array]
            return obj_array
        except ValueError:
            print("Cannot convert input to array")


class ToBase64(ValueTransformer):
    """A value transformer for expected base 64 values"""

    @staticmethod
    def transform(obj):
        try:
            return base64.b64encode(obj.encode('ascii')).decode('ascii')
        except ValueError:
            print("Cannot convert input to base64")


class ToFilePath(ValueTransformer):
    """A value transformer for expected file paths"""

    @staticmethod
    def transform(obj):
        try:
            return obj[0:len(obj) - len(re.split(r'[\\/]', obj)[-1])]
        except ValueError:
            print("Cannot convert input to path string")


class ToDirectoryPath(ValueTransformer):
    """A value transformer for expected directory path"""

    @staticmethod
    def transform(obj):
        try:
            return os.path.dirname(obj) + os.path.basename(obj)
        except ValueError:
            print("Cannot convert input to directory path string")


class ToFileName(ValueTransformer):
    """A value transformer for expected file names"""

    @staticmethod
    def transform(obj):
        try:
            return re.split(r'[\\/]', obj)[-1]
        except ValueError:
            print("Cannot convert input to file name")


class ToDomainName(ValueTransformer):
    """A value transformer for expected domain name"""

    @staticmethod
    def transform(url):
        try:
            if url is None:
                return
            parsed_url = urlparse(url)
            domain_name = parsed_url.netloc
            return domain_name
        except ValueError:
            print("Cannot convert input to domain name")


class ToIPv4(ValueTransformer):
    """A value transformer for converting an unsigned long to IPv4 string"""

    @staticmethod
    def transform(value):
        try:
            return socket.inet_ntoa((value & 0xffffffff).to_bytes(4, "big"))
        except ValueError:
            print("Cannot convert input to IPv4 string")


class DateTimeToUnixTimestamp(ValueTransformer):
    """A value transformer for converting python datetime object to Unix (millisecond) timestamp"""

    @staticmethod
    def transform(obj):
        try:
            return int((obj - datetime(1970, 1, 1)).total_seconds() * 1000)
        except ValueError:
            print("Cannot convert input {} to Unix timestamp".format(obj))


class NaiveToUTC(tzinfo):
    """
        A class for converting naive datetime object to UTC timezone datetime object
        naive datetime: "2017-09-29T15:00:00
        datetime with timezone offset: "2017-09-29T15:00:00+0300
        This class is declared to help convert Naive datetime format to datetime format with timezone(UTC here)
        """
    _dst = timedelta(0)

    def tzname(self, **kwargs):
        return "UTC"

    def utcoffset(self, dt):
        return self.__class__._dst

    def dst(self, dt):
        return self.__class__._dst


class TimestampToUTC(ValueTransformer):
    """
    A value transformer for converting a UTC timestamp (YYYY-MM-DDThh:mm:ss.000Z)
    to %d %b %Y %H:%M:%S %z"(23 Oct 2018 12:20:14 +0000)
    """

    @staticmethod
    def transform(timestamp, is_default=False):
        """
        Transformer for converting naive or non-naive to UTC time
        :param timestamp: datetime.datetime object(datetime.datetime(2019, 8, 22, 15, 44, 11, 716805)) or
                            '2019-07-25T10:43:10.003Z'
        :param is_default: True if timestamp like '2019-07-25T10:43:10.003Z'
                            False if timestamp like datetime.datetime(2019, 8, 22, 15, 44, 11, 716805)
        :return: str, e.g. : 25 Jul 2019 10:43:10 +0000
        """
        if re.search(r"\d{4}(-\d{2}){2}T\d{2}(:\d{2}){2}Z", str(timestamp)):
            input_time_pattern = '%Y-%m-%dT%H:%M:%SZ'
        else:
            input_time_pattern = '%Y-%m-%dT%H:%M:%S.%fZ'
        output_time_pattern = '%d %b %Y %H:%M:%S %z'
        if not is_default:
            # convert timestamp to datetime object
            datetime_obj = datetime.strptime(timestamp, input_time_pattern)
        else:
            datetime_obj = timestamp
        converted_time = datetime.strftime(datetime_obj.replace(tzinfo=NaiveToUTC()), output_time_pattern)
        return converted_time


class SetToOne(ValueTransformer):
    """Send back integer = 1 irrespective of the obj"""

    @staticmethod
    def transform(obj):
        try:
            return int("1")
        except ValueError:
            print("Cannot convert input {} to integer".format(obj))

# TODO: rename classes to be more generic since they can be reused by other data sources


class EpochToGuardium(ValueTransformer):
    """A value transformer for Epoch to Guardium timestamp"""

    @staticmethod
    def transform(epoch):
        try:
            return (datetime.fromtimestamp(int(epoch) / 1000, timezone.utc).strftime('%Y-%m-%d %H:%M:%S'))
        except ValueError:
            print("Cannot convert epoch value {} to timestamp".format(epoch))


class GuardiumToTimestamp(ValueTransformer):
    """A value transformer for converting Guardium timestamp to regular timestamp"""

    @staticmethod
    def transform(gdmTime):
        rgx = r"(\d\d\d\d-\d\d-\d\d)\s(\d\d:\d\d:\d\d)"
        mtch = (re.findall(rgx, gdmTime))[0]
        return (mtch[0] + 'T' + mtch[1]) + '.000Z'


class TimestampToGuardium(ValueTransformer):
    """A value transformer for converting  regular timestamp to Guardium timestamp"""

    @staticmethod
    def transform(timestamp):
        rgx = r"(\d\d\d\d-\d\d-\d\d).(\d\d:\d\d:\d\d)"
        mtch = (re.findall(rgx, timestamp))[0]
        return (mtch[0] + ' ' + mtch[1])


class Ymd_HMSToTimestamp(ValueTransformer):
    """A value transformer for the timestamps but adds ONE second to the time stamp.  Reason: Use when modified date is missing"""

    @staticmethod
    def transform(dt_Str):
        dt_obj = datetime.strptime(dt_Str, '%Y-%m-%d %H:%M:%S')
        dt_objOne = dt_obj + timedelta(seconds=1)
        return (datetime.fromtimestamp(datetime.timestamp(dt_objOne), timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z')


def get_all_transformers():
    return {"SplunkToTimestamp": SplunkToTimestamp, "EpochToTimestamp": EpochToTimestamp, "ToInteger": ToInteger, "ToString": ToString,
            "ToLowercaseArray": ToLowercaseArray, "ToBase64": ToBase64, "ToFilePath": ToFilePath, "ToFileName": ToFileName,
            "StringToBool": StringToBool, "ToDomainName": ToDomainName, "TimestampToMilliseconds": TimestampToMilliseconds,
            "EpochSecondsToTimestamp": EpochSecondsToTimestamp, "ToIPv4": ToIPv4,
            "DateTimeToUnixTimestamp": DateTimeToUnixTimestamp, "NaiveTimestampToUTC": TimestampToUTC,
            "ToDirectoryPath": ToDirectoryPath, "MsatpToTimestamp": MsatpToTimestamp, "FormatTCPProtocol": FormatTCPProtocol,
            "MsatpToRegistryValue": MsatpToRegistryValue, "FormatMac": FormatMac,
            "SetToOne": SetToOne, "Ymd_HMSToTimestamp": Ymd_HMSToTimestamp, "TimestampToGuardium": TimestampToGuardium,
            "GuardiumToTimestamp": GuardiumToTimestamp, "EpochToGuardium": EpochToGuardium, "AwsToTimestamp": AwsToTimestamp}




from stix_shifter_utils.modules.base.stix_translation.base_translator import BaseTranslator
from stix_shifter_utils.stix_translation.src.json_to_stix.json_to_stix import JSONToStix
from stix_shifter_utils.modules.base.stix_translation.base_query_translator import BaseQueryTranslator
from os import path
import json


class CARToStix(JSONToStix):

  def __init__(self, dialect=None, callback=None):
    basepath = path.dirname(__file__)
    filepath = path.abspath(
        path.join(basepath, "json", "to_stix_map.json"))
    self.default_mapping_file_path = filepath
    self.callback=None

  def translate_results(self, data_source, data, options, mapping=None):
      """
      Translates JSON data into STIX results based on a mapping file
      :param data: JSON formatted data to translate into STIX format
      :type data: str
      :param mapping: The mapping file path to use as instructions on how to translate the given JSON data to STIX. Defaults the path to whatever is passed into the constructor for JSONToSTIX (This should be the to_stix_map.json in the module's json directory)
      :type mapping: str (filepath)
      :return: STIX formatted results
      :rtype: str
      """

      json_data = json.loads(data)
      for obj in json_data:
        typ = obj.pop('object', '')
        fields = obj.pop('fields', [])
        for field in fields:
          obj[f"{typ}.{field}"] = fields[field]

      return JSONToStix.translate_results(self, data_source, json.dumps(json_data), options, mapping)




from abc import ABCMeta, abstractmethod


class BaseDeleteConnector(object, metaclass=ABCMeta):
    @abstractmethod
    def delete_query_connection(self, search_id):
        """
        Deletes a query from the specified datasource

        Args:
            search_id (str): The datasource query ID.

        Returns:
            dict: The return value.
                keys:
                    success (bool): True or False
                    error (str): error message (when success=False)
        """
        pass



from stix_shifter_utils.stix_translation.src.patterns.pattern_objects import StartStopQualifier, ObservationExpression, \
    ComparisonExpression, Pattern, CombinedComparisonExpression, CombinedObservationExpression
from stix_shifter_utils.stix_translation.src.utils.exceptions import DataMappingException


class UnmappedAttributeStripper:

    def __init__(self, antlr_object, data_model_mapping):
        self.dmm = data_model_mapping
        self.unmapped_attributes = []
        self.transformed_parsing = self._traverse_parsing_tree(antlr_object)

    def _traverse_combined_expression(self, root):
        expression1 = self._traverse_parsing_tree(root.expr1)
        expression2 = self._traverse_parsing_tree(root.expr2)

        if (not isinstance(root, CombinedObservationExpression) and
            (expression1 == "delete" or expression2 == "delete") and
                root.operator.name == "And"):
            return "delete"

        if (expression1 == "delete" and expression2 == "delete"):
            return "delete"

        if (expression1 == "delete"):
            return expression2

        if (expression2 == "delete"):
            return expression1

        root.expr1 = expression1
        root.expr2 = expression2

        return root

    def _parse_pattern_expression(self, root):
        expression = self._traverse_parsing_tree(root.expression)
        if (expression == "delete"):
            expression = None
        root.expression = expression
        return root

    def _parse_start_stop_qualifier(self, root):
        expression = self._traverse_parsing_tree(root.observation_expression)
        if (expression == "delete"):
            return expression
        root.observation_expression = expression
        return root

    def _parse_observation_expression(self, root):
        if (root.comparison_expression):
            expression = self._traverse_parsing_tree(root.comparison_expression)
            if (expression == "delete"):
                return expression
            root.comparison_expression = expression
        return root

    def _parse_comparison_expression(self, root):
        stix_object, stix_field = root.object_path.split(':')
        mapped_fields_array = self.dmm.map_field(stix_object, stix_field)
        if not mapped_fields_array:
            self.unmapped_attributes.append(root.object_path)
            return "delete"
        else:
            return root

    def _traverse_parsing_tree(self, root):
        if root:
            if isinstance(root, Pattern):
                root = self._parse_pattern_expression(root)
            if isinstance(root, StartStopQualifier):
                root = self._parse_start_stop_qualifier(root)
            if isinstance(root, CombinedObservationExpression):
                return self._traverse_combined_expression(root)
            if isinstance(root, ObservationExpression):
                root = self._parse_observation_expression(root)
            if isinstance(root, CombinedComparisonExpression):
                return self._traverse_combined_expression(root)
            if isinstance(root, ComparisonExpression):
                root = self._parse_comparison_expression(root)
            return root


def strip_unmapped_attributes(antlr_parsing, data_model_mapping=None):
    if not data_model_mapping:
        return antlr_parsing
    attribute_stripper = UnmappedAttributeStripper(antlr_parsing, data_model_mapping)
    modified_antlr_parsing = attribute_stripper.transformed_parsing
    if not modified_antlr_parsing.expression:
        modified_antlr_parsing = None
    return {"parsing": modified_antlr_parsing, "unmapped_stix": attribute_stripper.unmapped_attributes}


from enum import Enum

class ErrorMapperBase():
    @staticmethod
    def set_error_code(return_obj, code, message=None):
        if isinstance(code, Enum):
            return_obj['code'] = code.value
        else:
            return_obj['code'] = code
        if message is not None:
            return_obj['error'] = message